__all__: list[str]

class RankWarning(UserWarning): ...

def trimseq(seq): ...
def as_series(alist, trim=...): ...
def trimcoef(c, tol=...): ...
def getdomain(x): ...
def mapparms(old, new): ...
def mapdomain(x, old, new): ...
